(ns document-storage.git.core
  (:require [clj-jgit.porcelain :as jp]
            [clojure.tools.logging :as log]
            [clojure.java.io :as io]
            [clojure.edn :refer [read-string]]
            [document-storage.git.query :refer [query-fits? update-queries update-queries-remove-document]]
            [document-storage.git.versioning :as versioning]
            [document-storage.git.repository :refer [repositories init-repository valid-id?
                                                 default-repository
                                                 no-commit-repository?
                                                 cached-repository?
                                                 get-file-path
                                                 get-repo
                                                 get-repo-data
                                                 get-repository-path]]
            [document-storage.git.views :refer [get-views generate-views]]
            [fipp.edn :refer [pprint] :rename {pprint fipp}]
            [document-storage.protocol :as protocol]))

(defonce repo-dir (atom ""))

(defonce upstream-ssh-identity "upstream")
(defonce storage-directory (atom ""))
(defonce cache (atom {}))


(defn load-config [dir]
  (-> dir
      (str "/config.edn")
      slurp
      read-string))


(defn reset-cache! []
  (reset! cache {}))

(defn- commit-document-change [message repository]
  (jp/git-commit (get-repo repository) message
                 :committer {:name "Document-Manager" :email "burke@fh-muenster.de"}))

(defn cache-document [repository id doc]
  (swap! cache update-in [repository :documents] assoc id doc)
  doc)

(defn cache-query [repository query result]
  (swap! cache update-in [repository :queries] assoc query (into #{} (keys result)))
  result)


(defn uncache-document [repository id]
  (swap! cache update-in [repository :queries] update-queries-remove-document id)
  (swap! cache update-in [repository :documents] dissoc id))

(defn update-cache-document [repository id document]
  (swap! cache update-in [repository :queries] update-queries id document)
  (swap! cache assoc-in [repository :documents id] document))




(defn get-upstreams [repo-name]
  (-> (get-repo-data repo-name)
      :upstreams))

(defn init-upstreams [repo-name]
  (let [git (get-repo repo-name)
        upstreams (get-upstreams :default)]
    (doseq [[id {:keys [url]}] upstreams]
      (jp/git-remote-remove git (name id))
      (jp/git-remote-add git (name id) url)
      )
    )
  )

(defn push-upstreams [repo-name]
  (let [git (get-repo repo-name)
        upstreams (jp/git-remote-list git)]
    (jp/with-identity {:name (str @storage-directory "/" upstream-ssh-identity) :exclusive true}
      (doseq [[id urls] upstreams]
        (jp/git-push git :remote id)
        ))
    )
  )





(defrecord git-storage [options]
  protocol/storage
  (init-storage [{{:keys [dir] :as options} :options :as this}]
    (println options)
    (log/info "Init Storage in " dir)
    (reset! storage-directory dir)
    (if-let [config (load-config dir)]
      (do
        (doseq [repo (keys (:repositories config))]
          (swap! repositories
                 assoc repo
                 (merge {:directory (str dir "/" (name repo))
                         :unsaved-files []
                         :git-repo (init-repository (str dir "/" (name repo)))}
                        (get (:repositories config) repo))))
        (update this :options
                assoc
                :repositories @repositories
                :cache (atom {})))
      (log/error "Unable to initialize Document Storage - no config!")))

  (load-document [this id repository]
    (let [cached-repository (cached-repository? repository)]
      (or
       (and cached-repository (get-in @cache [repository :documents id]))
       (when (valid-id? id repository)
         (let [result (read-string (slurp (get-file-path id repository)))]
           (generate-views repository id result)
           (if cached-repository
             (cache-document repository id result)
             result
             ))))))

  (get-document-size [this id repository]
    (when (valid-id? id repository)
      (let [file (io/file (get-file-path id repository))]
        (when (.isFile file)
          (.length file)))))

  (save-document [this id document repository]
    (let [file-path (get-file-path id repository)
          document-str (with-out-str (fipp document))]
      (when-not (and (valid-id? id repository)
                     (= (slurp file-path) document-str))
        (update-cache-document repository id document)
        (generate-views repository id document)
        (with-open [w (io/writer file-path)]
          (.write w document-str))
        (when-not (no-commit-repository? repository)
          (jp/git-add (get-repo repository) id)
          (commit-document-change (str "saved new version of " id) repository)))))

  (delete-document [this id repository]
    (when (valid-id? id repository)
      (uncache-document repository id)
      (io/delete-file (get-file-path id :units) true)
      (jp/git-rm (get-repo repository) id)
      (commit-document-change (str "deleted document " id) repository)))

  (list-documents [this repository]
    (let [repo-path (get-repository-path repository)
          files     (->> (io/file repo-path)
                         .listFiles
                         (filter #(.isFile %))
                         (map #(.getName %))
                         )]
      files))

  (find-documents [this query repository]
    (let [docs (.list-documents this :repository repository)
          cached-repository (cached-repository? repository)]
      (or (when cached-repository
            (when-let [cached-query (get-in @cache [repository :queries query])]
              (into {} (map (fn [fname]
                              (let [doc (.load-document this fname :repository repository)]
                                [fname doc]))
                            cached-query))))
          (let [result (into {} (keep (fn [fname]
                                        (let [doc (.load-document this fname :repository repository)]
                                          (when (query-fits? doc query)
                                            [fname doc]
                                            )
                                          ))
                                      docs))]
            (if cached-repository
              (cache-query repository query result)
              result)))))
  (shutdown [this]
    (pmap (fn [[id repo]]
            (push-upstreams id)
            ) @repositories))

  (list-versions [this id repository]
    (let [repo (get-repo repository)]
      (versioning/list-commits repo id)))

  (load-version [this id version repository]
    (let [versions (.list-versions this id :repository repository)
          repo (get-repo repository)]
      (when-let [commit (-> (filter #(= (:version %) version) versions)
                            first :commit)]
        (versioning/load-revision commit repo id))))

  )

